"""
Different utility functions practical for ROV control
"""

import ctypes
import os
import platform
import signal
import socket
import struct
import subprocess
import warnings


def detect_pi():
    return platform.linux_distribution()[0].lower() == 'debian'


if detect_pi():
    import serial
    import fcntl


def is_int(number):
    if isinstance(number, int):
        return True
    else:
        try:
            if isinstance(int(number), int):
                return True
        except ValueError:
            pass
    return False


def resolution_to_tuple(resolution):
    if 'x' not in resolution:
        raise ValueError('Resolution must be in format WIDTHxHEIGHT')
    screen_size = tuple([int(val) for val in resolution.split('x')])
    if len(screen_size) is not 2:
        raise ValueError('Error in parsing resolution, len is not 2')
    return screen_size


def preexec_function():
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def valid_resolution(resolution):
    if 'x' in resolution:
        w, h = resolution.split('x')
        if is_int(w) and is_int(h):
            return resolution
    warning('Resolution must be WIDTHxHEIGHT')


def server_ip(port):
    online_ips = []
    for interface in [b'eth0', b'wlan0']:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            ip = socket.inet_ntoa(fcntl.ioctl(
                sock.fileno(),
                0x8915,
                struct.pack('256s', interface[:15])
            )[20:24])
            online_ips.append(ip)
        except OSError:
            pass
        sock.close()
    return ' or '.join(['{}:{}'.format(ip, port) for ip in online_ips])


def check_requirements():
    if detect_pi():
        camera = subprocess.check_output(['vcgencmd',
                                          'get_camera']).decode().rstrip()
        if '0' in camera:
            warning('Camera not enabled or connected properly')
            return False
        else:
            return True
    else:
        warning('eduROV only works on a raspberry pi')
        return False


def send_arduino(msg, serial_connection):
    """
    Send the *msg* over the *serial_connection*

    Adds a hexadecimal number of 6 bytes to the start of the message before
    sending it. "hello there" -> "0x000bhello there"

    Parameters
    ----------
    msg : str or bytes
        the message you want to send
    serial_connection : object
        the :code:`serial.Serial` object you want to use for sending
    """
    if not isinstance(msg, bytes):
        msg = str(msg).encode()
    length = "{0:#0{1}x}".format(len(msg), 6).encode()
    data = length + msg
    serial_connection.write(data)


def receive_arduino(serial_connection):
    """
    Returns a message received over *serial_connection*

    Expects that the message received starts with a 6 bytes long number
    describing the size of the remaining data. "0x000bhello there" -> "hello
    there".

    Parameters
    ----------
    serial_connection : object
        the :code:`serial.Serial` object you want to use for receiving

    Returns
    -------
    msg : str or None
        the message received or None
    """
    if serial_connection.inWaiting():
        msg = serial_connection.readline().decode().rstrip()
        if len(msg) >= 6:
            try:
                length = int(msg[:6], 0)
                data = msg[6:]
                if length == len(data):
                    return data
                else:
                    warning('Received incomplete serial string: {}'
                            .format(data), 'default')
            except ValueError:
                pass
    return None


def send_arduino_simple(msg, serial_connection):
    """
    Send the *msg* over the *serial_connection*

    Same as :code:`send_arduino`, but doesn't add anything to the message
    before sending it.

    Parameters
    ----------
    msg : str or bytes
        the message you want to send
    serial_connection : object
        the :code:`serial.Serial` object you want to use for sending
    """
    if not isinstance(msg, bytes):
        msg = str(msg).encode()
    serial_connection.write(msg)


def receive_arduino_simple(serial_connection, min_length=1):
    """
    Returns a message received over *serial_connection*

    Same as :code:`receive_arduino` but doesn't expect that the message starts
    with a hex number.

    Parameters
    ----------
    serial_connection : object
        the :code:`serial.Serial` object you want to use for receiving
    min_length : int, optional
        if you only want that the function to only return the string if it is
        at least this long.

    Returns
    -------
    msg : str or None
        the message received or None
    """
    if serial_connection.inWaiting():
        msg = serial_connection.readline().decode().rstrip()
        if len(msg) >= min_length:
            return msg
        else:
            return None


def serial_connection(port='/dev/ttyACM0', baudrate=115200, timeout=0.05):
    """
    Establishes a serial connection

    Parameters
    ----------
    port : str, optional
        the serial port you want to use
    baudrate : int, optional
        the baudrate of the serial connection
    timeout : float, optional
        read timeout value

    Returns
    -------
    connection : class or None
        a :code:`serial.Serial` object if successful or None if not
    """
    try:
        ser = serial.Serial(port, baudrate, timeout=timeout)
        ser.close()
        ser.open()
        return ser
    except FileNotFoundError:
        pass
    except serial.serialutil.SerialException:
        pass
    except ValueError:
        pass
    warning(message="""Could not establish serial connection at {}\n
    Try running 'ls /dev/*tty*' to find correct port"""
            .format(port), filter='default')
    return None


def warning(message, filter='error', category=UserWarning):
    warnings.simplefilter(filter, category)
    warnings.formatwarning = warning_format
    warnings.warn(message)


def warning_format(message, category, filename, lineno,
                   file=None, line=None):
    return 'WARNING:\n  {}: {}\n  File: {}:{}\n'.format(
        category.__name__, message, filename, lineno)


def free_drive_space(as_string=False):
    """
    Checks and returns the remaining free drive space

    Parameters
    ----------
    as_string : bool, optional
        set to True if you want the function to return a formatted string.
        4278 -> 4.28 GB

    Returns
    -------
    space : float or str
        the remaining MB in float or as string if *as_string=True*
    """
    if platform.system() == 'Windows':
        free_bytes = ctypes.c_ulonglong(0)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p('/'),
                                                   None, None,
                                                   ctypes.pointer(free_bytes))
        mb = free_bytes.value / 1024 / 1024
    else:
        st = os.statvfs('/')
        mb = st.f_bavail * st.f_frsize / 1024 / 1024

    if as_string:
        if mb >= 1000:
            return '{:.2f} GB'.format(mb / 1000)
        else:
            return '{:.0f} MB'.format(mb)
    else:
        return mb


def cpu_temperature():
    """
    Checks and returns the on board CPU temperature

    Returns
    -------
    temperature : float
        the temperature
    """
    cmds = ['/opt/vc/bin/vcgencmd', 'measure_temp']
    response = subprocess.check_output(cmds).decode()
    return float(response.split('=')[1].split("'")[0].rstrip())



import argparse

from edurov.utils import valid_resolution, check_requirements
from .start import main


def edurov_web(args=None):
    parser = argparse.ArgumentParser(
        description='Start the Engage eduROV web server')
    parser.add_argument(
        '-r',
        metavar='RESOLUTION',
        type=str,
        default='1024x768',
        help='''resolution, use format WIDTHxHEIGHT (default 1024x768)''')
    parser.add_argument(
        '-fps',
        metavar='FRAMERATE',
        type=int,
        default=30,
        help='framerate for the camera (default 30)')
    parser.add_argument(
        '-port',
        metavar='SERVER_PORT',
        type=int,
        default=8000,
        help="which port the server should serve it's content (default 8000)")
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        help='set to print debug information')

    args = parser.parse_args()

    if check_requirements():
        video_res = valid_resolution(args.r)
        main(
            video_resolution=video_res,
            fps=args.fps,
            server_port=args.port,
            debug=args.debug)



import os
import subprocess
import time
from multiprocessing import Process

from edurov.sync import start_sync_classes
from edurov.utils import warning, preexec_function, detect_pi
from edurov.web import start_http_server

if detect_pi():
    import Pyro4

class WebMethod(object):
    """
    Starts a video streaming from the rasparry pi and a webserver that can
    handle user input and other requests.

    Parameters
    ----------
    index_file : str
        absolute path to the frontpage of the webpage, must be called
        ``index.html``
    video_resolution : str, optional
        a string representation of the wanted video resolution in the format
        WIDTHxHEIGHT
    fps : int, optional
        wanted framerate, may not be achieved depending on available resources
        and network
    server_port : int, optional
        the web page will be served at this port
    debug : bool, optional
        if set True, additional information will be printed for debug
        purposes
    runtime_functions : callable or list, optional
        should be a callable function or a list of callable functions, will be
        started as independent processes automatically
    custom_response : callable, optional
        if set, this function will be called if default web server is not able
        to handle a GET request, should return a str or None. If returned value
        starts with ``redirect=`` followed by a path, the browser wil redirect
        the user to this path. The callable must accept two parameters whereas
        the second one is the requested path
    """
    def __init__(self, index_file, video_resolution='1024x768', fps=30,
                 server_port=8000, debug=False, runtime_functions=None,
                 custom_response=None):

        self.res = video_resolution
        self.fps = fps
        self.server_port = server_port
        self.debug = debug
        self.run_funcs = self._valid_runtime_functions(runtime_functions)
        self.cust_resp = self._valid_custom_response(custom_response)
        self.index_file = self._valid_index_file(index_file)

    def _valid_custom_response(self, custom_response):
        if custom_response:
            if not callable(custom_response):
                warning('custom_response parameter has to be a callable '
                        'function, not type {}'.format(type(custom_response)))
                return None
        return custom_response

    def _valid_runtime_functions(self, runtime_functions):
        if runtime_functions:
            if callable(runtime_functions):
                runtime_functions = [runtime_functions]
            elif isinstance(runtime_functions, list):
                for f in runtime_functions:
                    if not callable(f):
                        warning(
                            'Parameter runtime_functions has to be a function '
                            'or a list of functions, not {}'.format(type(f)))
            else:
                warning('Parameter runtime_functions has to be a function '
                        'or a list of functions, not {}'
                        .format(type(runtime_functions)))
        return runtime_functions

    def _valid_index_file(self, file_path):
        if not 'index.html' in file_path:
            warning('The index files must be called "index.html')
        if os.path.isfile(file_path):
            return os.path.abspath(file_path)
        else:
            warning('could not find "{}", needs absolute path'
                    .format(file_path))
        return None

    def serve(self, timeout=None):
        """
        Will start serving the web page defined by the index_file parameter

        Parameters
        ----------
        timeout : int, optional
            if set, the web page will only be served for that many seconds
            before it automatically shuts down
        """
        start = time.time()
        name_server = subprocess.Popen('pyro4-ns', shell=False,
                                       preexec_fn=preexec_function)
        time.sleep(2)
        pyro_classes = Process(target=start_sync_classes)
        pyro_classes.start()
        time.sleep(4)
        web_server = Process(
            target=start_http_server,
            args=(self.res, self.fps, self.server_port, self.index_file,
                  self.debug, self.cust_resp))
        web_server.daemon = True
        web_server.start()
        processes = []
        if self.run_funcs:
            for f in self.run_funcs:
                p = Process(target=f)
                p.daemon = True
                p.start()
                processes.append(p)

        with Pyro4.Proxy("PYRONAME:ROVSyncer") as rov:
            try:
                while rov.run:
                    if timeout:
                        if time.time() - start >= timeout:
                            break
            except KeyboardInterrupt:
                pass
            finally:
                print('Shutting down')
                web_server.terminate()
                rov.run = False
                if self.run_funcs:
                    for p in processes:
                        p.join(3)
                pyro_classes.terminate()
                name_server.terminate()



"""
Synchronizing the state of ROV and controller
"""

import os
import time

import Pyro4


class Key(object):
    """Manages the state of a specific key on the keyboard"""

    def __init__(self, KeyASCII, ASCII, common, keycode, mode='hold'):
        self.state = False
        self.KeyASCII = KeyASCII
        self.ASCII = ASCII
        self.common = common
        self.mode = mode
        if keycode:
            self.keycode = int(keycode)
        else:
            self.keycode = None

    def keydown(self):
        if self.mode == 'toggle':
            self.state = not self.state
        else:
            self.state = True

    def keyup(self):
        if self.mode != 'toggle':
            self.state = False

    def __str__(self):
        return str(vars(self))


@Pyro4.expose
class KeyManager(object):
    """Keeps control of all user input from keyboard"""

    def __init__(self):
        self.keys = {}
        cwd = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(cwd, 'keys.txt'), 'r') as f:
            for line in f.readlines()[1:]:
                KeyASCII = line[0:14].rstrip()
                ASCII = line[14:22].rstrip()
                common = line[22:44].rstrip()
                keycode = line[44:].rstrip()
                if keycode:
                    dict_key = int(keycode)
                else:
                    dict_key = KeyASCII
                self.keys.update({
                    dict_key: Key(KeyASCII, ASCII, common, keycode)})

    def set_mode(self, key, mode):
        """
        Set the press mode for the key to *hold* or *toggle*

        Parameters
        ----------
        key : int or str
            key identifier as described above
        mode : str
            *hold* or *toggle*
        """
        self._get(key).mode = mode

    def set(self, key, state):
        """
        Set the state of the key to True or False

        Parameters
        ----------
        key : int or str
            key identifier as described above
        state : bool
            *True* or *False*
        """
        self._get(key).state = bool(state)

    def _get(self, key_idx, make_exception=True):
        """
        Returns the Key object identified by *key_idx*

        Parameters
        ----------
        key_idx : int or str
            key identifier as described above
        make_exception : bool, optional
            As default an exception is raised if the key is not found, this
            behavior can be changed be setting it to *False*

        Returns
        -------
        key : Key object
            list items is *namedtuple* of type *LiItem*
        """
        if key_idx in self.keys:
            return self.keys[key_idx]
        elif isinstance(key_idx, str):
            for dict_key in self.keys:
                if key_idx in [self.keys[dict_key].common,
                               self.keys[dict_key].KeyASCII]:
                    return self.keys[dict_key]
        if make_exception:
            raise ValueError('Could not find key {}'.format(key_idx))
        else:
            return None

    def state(self, key):
        """
        Returns the state of *key*

        Parameters
        ----------
        key : int or str
            key identifier as described above

        Returns
        -------
        state : bool
            *True* or *False*
        """
        return self._get(key).state

    def keydown(self, key, make_exception=False):
        """
        Call to simulate a keydown event

        Parameters
        ----------
        key : int or str
            key identifier as described above
        make_exception : bool, optional
            As default an exception is raised if the key is not found, this
            behavior can be changed be setting it to *False*
        """
        btn = self._get(key, make_exception=make_exception)
        if btn:
            btn.keydown()

    def keyup(self, key, make_exception=False):
        """
        Call to simulate a keyup event

        Parameters
        ----------
        key : int or str
            key identifier as described above
        make_exception : bool, optional
            As default an exception is raised if the key is not found, this
            behavior can be changed be setting it to *False*
        """
        btn = self._get(key, make_exception=make_exception)
        if btn:
            btn.keyup()

    @property
    def qweasd_dict(self):
        """
        Dictionary with the state of the letters q, w, e, a, s and d
        """
        state = {
            'q': self._get(81).state,
            'w': self._get(87).state,
            'e': self._get(69).state,
            'a': self._get(65).state,
            's': self._get(83).state,
            'd': self._get(68).state,
        }
        return state

    @property
    def arrow_dict(self):
        """
        Dictionary with the state of the keys *up arrow*, *down arrow*,
        *left arrow* and *right arrow*
        """
        state = {
            'up arrow': self._get(38).state,
            'down arrow': self._get(40).state,
            'left arrow': self._get(37).state,
            'right arrow': self._get(39).state,
        }
        return state


@Pyro4.expose
class ROVSyncer(object):
    """Holds all variables for ROV related to control and sensors"""

    def __init__(self):
        self._sensor = {'time': time.time()}
        self._actuator = {}
        self._run = True

    @property
    def sensor(self):
        """
        Dictionary holding sensor values

        :getter: Returns sensor values as dict
        :setter: Update sensor values with dict
        :type: dict
        """
        return self._sensor

    @sensor.setter
    def sensor(self, values):
        self._sensor.update(values)
        self._sensor['time'] = time.time()

    @property
    def actuator(self):
        """
        Dictionary holding actuator values

        :getter: Returns actuator values as dict
        :setter: Update actuator values with dict
        :type: dict
        """
        return self._actuator

    @actuator.setter
    def actuator(self, values):
        self._actuator.update(values)
        self._actuator['time'] = time.time()

    @property
    def run(self):
        """
        Bool describing if the ROV is still running

        :getter: Returns bool describing if the ROV is running
        :setter: Set to False if the ROV should stop
        :type: bool
        """
        return self._run

    @run.setter
    def run(self, bool_):
        self._run = bool_


def start_sync_classes():
    """Registers pyro classes in name server and starts request loop"""
    rov = ROVSyncer()
    keys = KeyManager()
    with Pyro4.Daemon() as daemon:
        rov_uri = daemon.register(rov)
        keys_uri = daemon.register(keys)
        with Pyro4.locateNS() as ns:
            ns.register("ROVSyncer", rov_uri)
            ns.register("KeyManager", keys_uri)
        daemon.requestLoop()


if __name__ == "__main__":
    start_sync_classes()
